"""
* Omnitils
* A collection of generic, useful utilities for the Python super-fan on a tight schedule.
* Copyright (c) Hexproof Systems <hexproofsystems@gmail.com>
* LICENSE: Mozilla Public License 2.0
"""
__version__ = '1.4.0'
